from abc import ABC, abstractmethod
from typing import Dict

from radarpipeline.datalib import Data
from radarpipeline.datatypes.data_types import DataType


class DataReader(ABC):
    """
    Abstract class for reading the RADAR data
    """

    config: Dict

    def __init__(self, config: Dict) -> None:
        self.config = config

    @abstractmethod
    def read_data(self) -> Data:
        pass


class SchemaReader(ABC):
    """
    Abstract class for reading the RADAR data schema
    """

    schema_dir: str
    schema_file: str

    def __init__(self, schema_dir: str) -> None:
        self.schema_dir = schema_dir

    @abstractmethod
    def is_schema_present() -> bool:
        pass

    @abstractmethod
    def get_schema(self):
        pass


class DataWriter(ABC):
    """
    Abstract class for writing the RADAR data
    """

    features: Dict[str, DataType]
    output_dir: str

    def __init__(self, features: Dict[str, DataType], output_dir: str) -> None:
        self.features = features
        self.output_dir = output_dir

    @abstractmethod
    def write_data(self) -> None:
        pass



import json
import logging
import os
from typing import Any, Dict, List, Optional, Union

import pyspark.sql as ps
from pyspark.sql import SparkSession
from pyspark.sql.types import StructField, StructType

from radarpipeline.common import constants
from radarpipeline.datalib import RadarData, RadarUserData, RadarVariableData
from radarpipeline.io.abc import DataReader, SchemaReader

from multiprocessing import Pool
from functools import partial
from datetime import datetime

logger = logging.getLogger(__name__)


class SparkCSVDataReader(DataReader):
    """
    Read CSV data from local directory using pySpark
    """

    def __init__(self, config: Dict, required_data: List[str], df_type: str = "pandas",
                 spark_config: Dict = {}):
        super().__init__(config)
        default_spark_config = {'spark.executor.instances': 6,
                                'spark.driver.memory': '10G',
                                'spark.executor.cores': 4,
                                'spark.executor.memory': '10g',
                                'spark.memory.offHeap.enabled': True,
                                'spark.memory.offHeap.size': '20g',
                                'spark.driver.maxResultSize': '0'}

        self.required_data = required_data
        self.df_type = df_type
        self.source_path = self.config['config'].get("source_path", "")
        self.spark_config = default_spark_config
        if spark_config is not None:
            self.spark_config.update(spark_config)
        self.spark = self._initialize_spark_session()

    def _initialize_spark_session(self) -> ps.SparkSession:
        """
        Initializes and returns a SparkSession

        Returns
        -------
        SparkSession
            A SparkSession object
        """

        """
        Spark configuration documentation:
        https://spark.apache.org/docs/latest/configuration.html

        `spark.executor.instances` is the number of executors to
        launch for an application.

        `spark.executor.cores` is the number of cores to =
        use on each executor.

        `spark.executor.memory` is the amount of memory to
        use per executor process.

        `spark.driver.memory` is the amount of memory to use for the driver process,
        i.e. where SparkContext is initialized, in MiB unless otherwise specified.

        `spark.memory.offHeap.enabled` is to enable off-heap memory allocation

        `spark.memory.offHeap.size` is the absolute amount of memory which can be used
        for off-heap allocation, in bytes unless otherwise specified.

        `spark.driver.maxResultSize` is the limit of total size of serialized results of
        all partitions for each Spark action (e.g. collect) in bytes.
        Should be at least 1M, or 0 for unlimited.
        """
        spark = (
            SparkSession.builder.master("local").appName("radarpipeline")
            .config('spark.executor.instances',
                    self.spark_config['spark.executor.instances'])
            .config('spark.executor.cores',
                    self.spark_config['spark.executor.cores'])
            .config('spark.executor.memory',
                    self.spark_config['spark.executor.memory'])
            .config('spark.driver.memory',
                    self.spark_config['spark.driver.memory'])
            .config('spark.memory.offHeap.enabled',
                    self.spark_config['spark.memory.offHeap.enabled'])
            .config('spark.memory.offHeap.size',
                    self.spark_config['spark.memory.offHeap.size'])
            .config('spark.driver.maxResultSize',
                    self.spark_config['spark.driver.maxResultSize'])
            .getOrCreate()
        )

        # Enable Apache Arrow for optimizations in Spark to Pandas conversion
        spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        # Fallback to use non-Arrow conversion in case of errors
        spark.conf.set("spark.sql.execution.arrow.pyspark.fallback.enabled", "true")
        # For further reading:
        # https://spark.apache.org/docs/3.0.1/sql-pyspark-pandas-with-arrow.html

        spark.sparkContext.setLogLevel("ERROR")
        logger.info("Spark Session created")
        return spark

    def close_spark_session(self):
        self.spark.stop()

    def read_data(self) -> RadarData:
        """
        Reads RADAR data from local CSV files

        Returns
        -------
        RadarData
            A RadarData object containing all the read data
        """

        user_data_dict = {}

        if not isinstance(self.source_path, list):
            self.source_path = [self.source_path]
        for source_path_item in self.source_path:
            for uid in os.listdir(source_path_item):
                # Skip hidden files
                if uid[0] == ".":
                    continue
                logger.info(f"Reading data for user: {uid}")
                variable_data_dict = {}
                for dirname in self.required_data:
                    if dirname not in os.listdir(os.path.join(source_path_item, uid)):
                        continue
                    logger.info(f"Reading data for variable: {dirname}")
                    absolute_dirname = os.path.abspath(
                        os.path.join(source_path_item, uid, dirname)
                    )
                    data_files = [
                        os.path.join(absolute_dirname, f)
                        for f in os.listdir(absolute_dirname)
                        if f.endswith(".csv.gz")
                    ]
                    schema = None
                    schema_reader = AvroSchemaReader(absolute_dirname)
                    if schema_reader.is_schema_present():
                        logger.info("Schema found")
                        schema = schema_reader.get_schema()
                    else:
                        logger.info("Schema not found, inferring from data file")
                    variable_data = self._read_variable_data_files(data_files, schema)
                    if variable_data.get_data_size() > 0:
                        variable_data_dict[dirname] = variable_data
                user_data_dict[uid] = RadarUserData(variable_data_dict, self.df_type)
        radar_data = RadarData(user_data_dict, self.df_type)
        return radar_data

    def _read_variable_data_files(
        self,
        data_files: List[str],
        schema: Optional[StructType] = None,
    ) -> RadarVariableData:
        """
        Reads data from a list of data files and returns a RadarVariableData object

        Parameters
        ----------
        data_files : List[str]
            List of data files to read
        schema : Optional[StructType]
            Schema to use for optimized reading

        Returns
        -------
        RadarVariableData
            A RadarVariableData object containing all the read data
        """
        if schema:
            df = self.spark.read.load(
                data_files,
                format="csv",
                header=True,
                schema=schema,
                inferSchema="false",
                encoding=constants.ENCODING,
            )
        else:
            df = self.spark.read.load(
                data_files,
                format="csv",
                header=True,
                inferSchema="true",
                encoding=constants.ENCODING,
            )

        if self.df_type == "pandas":
            df = df.toPandas()

        # To print the dataframe stats for cross-checking
        # print(df.show(5))

        variable_data = RadarVariableData(df, self.df_type)

        return variable_data


class AvroSchemaReader(SchemaReader):
    """
    Reads schema from local directory
    """

    def __init__(self, schema_dir: str) -> None:
        super().__init__(schema_dir)

    def is_schema_present(self) -> bool:
        """
        Checks if schema is present in local directory

        Returns
        -------
        bool
            True if schema is present, False otherwise
        """

        schema_dir_base = os.path.basename(self.schema_dir)
        self.schema_file = os.path.join(
            self.schema_dir, f"schema-{schema_dir_base}.json"
        )

        if os.path.exists(self.schema_file):
            self.schema_file = self.schema_file
            return True

        return False

    def get_schema(self) -> StructType:
        """
        Reads schema from local directory

        Returns
        -------
        StructType
            A StructType object defining the schema for pySpark
        """

        schema_fields = []
        schema_dict = json.load(
            open(
                os.path.join(self.schema_dir, self.schema_file),
                "r",
                encoding=constants.ENCODING,
            )
        )
        key_dict = schema_dict["fields"][0]
        value_dict = schema_dict["fields"][1]

        for key in key_dict["type"]["fields"]:
            name = key["name"]
            schema_fields.append(StructField(f"key.{name}", constants.STRING_TYPE))

        for value in value_dict["type"]["fields"]:
            name = value["name"]
            typ = value["type"]
            field_type = self._get_field(typ)
            schema_fields.append(StructField(f"value.{name}", field_type, True))

        schema = StructType(schema_fields)
        return schema

    def _get_field(self, data_type: Union[str, Dict, List]) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        data_type : Union[str, Dict]
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if type(data_type) is dict:
            spark_data_type = self._get_data_type_from_dict(data_type)
        elif type(data_type) is list:
            spark_data_type = self._get_superior_type_from_list(data_type)
        else:
            spark_data_type = self._get_data_type_from_mapping(data_type)

        return spark_data_type

    def _handle_unknown_data_type(self, data_type: Union[str, Dict, List]) -> Any:
        """
        Handles unknown data types

        Parameters
        ----------
        data_type : Union[str, Dict]
            Data type to handle

        Returns
        -------
        Any
            A Spark data type
        """

        logger.warning(f"Unknown data type: {data_type}. Returning String type.")
        return constants.STRING_TYPE

    def _get_data_type_from_mapping(self, data_type: Union[str, Dict, List]) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        data_type : str
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if data_type in constants.DATA_TYPE_MAPPING:
            spark_data_type = constants.DATA_TYPE_MAPPING[data_type]
        else:
            spark_data_type = self._handle_unknown_data_type(data_type)

        return spark_data_type

    def _get_data_type_from_dict(self, data_type: Dict) -> Any:
        """
        Returns a Spark data type for a given data type

        Parameters
        ----------
        data_type : Dict
            Data type to convert to a Spark data type

        Returns
        -------
        Any
            A Spark data type
        """

        if "type" in data_type:
            return self._get_field(data_type["type"])
        else:
            return self._handle_unknown_data_type(data_type)

    def _get_superior_type_from_list(self, data_type_list: List[Any]) -> Any:
        """
        Resolves a list data type to a Spark data type

        Parameters
        ----------
        data_type_list : List
            List data type to resolve

        Returns
        -------
        Any
            A Spark data type
        """

        spark_data_type_list = data_type_list.copy()

        if "null" in spark_data_type_list:
            spark_data_type_list.remove("null")

        for index, data_type in enumerate(spark_data_type_list):
            if type(data_type) is dict:
                spark_data_type_list[index] = self._get_data_type_from_dict(data_type)
            elif data_type in constants.DATA_TYPE_MAPPING:
                spark_data_type_list[index] = constants.DATA_TYPE_MAPPING[data_type]
            else:
                spark_data_type_list[index] = self._handle_unknown_data_type(data_type)

        if len(data_type_list) == 0:
            return constants.STRING_TYPE
        elif len(data_type_list) == 1:
            return spark_data_type_list[0]
        else:
            return self._get_superior_spark_type(spark_data_type_list)

    def _get_superior_spark_type(self, spark_data_type_list: List[Any]) -> Any:
        """
        Returns the superior Spark data type from a list of Spark data types

        Parameters
        ----------
        spark_data_type_list : List
            List of Spark data types

        Returns
        -------
        Any
            A Spark data type
        """

        spark_data_type_set = set(spark_data_type_list)

        # Conflicting types are datatypes which upon conversion can
        # lead to a loss of information
        are_conflicting_types = False
        are_integer_types = False
        are_float_types = False
        are_string_types = False
        are_boolean_types = False

        if len(spark_data_type_set.intersection(constants.INTEGER_TYPES)) > 0:
            are_integer_types = True

        if len(spark_data_type_set.intersection(constants.FLOATING_TYPES)) > 0:
            are_float_types = True

        if constants.STRING_TYPE in spark_data_type_set:
            are_string_types = True

        if constants.BOOLEAN_TYPE in spark_data_type_set:
            are_boolean_types = True

        if (
            int(are_integer_types)
            + int(are_float_types)
            + int(are_string_types)
            + int(are_boolean_types)
            > 1
        ):
            are_conflicting_types = True

        if not are_conflicting_types:
            if are_integer_types:
                if constants.LONG_TYPE in spark_data_type_set:
                    return constants.LONG_TYPE
                elif constants.INT_TYPE in spark_data_type_set:
                    return constants.INT_TYPE
                elif constants.SHORT_TYPE in spark_data_type_set:
                    return constants.SHORT_TYPE
                else:
                    return constants.BYTE_TYPE
            elif are_float_types:
                if constants.DOUBLE_TYPE in spark_data_type_set:
                    return constants.DOUBLE_TYPE
                else:
                    return constants.FLOAT_TYPE
            elif are_string_types:
                return constants.STRING_TYPE
            elif are_boolean_types:
                return constants.BOOLEAN_TYPE
        else:
            logger.warning(
                f"Conflicting types: {spark_data_type_list}. Returning String type."
            )
            return constants.STRING_TYPE


class Reader():
    '''
    Class for reading data from a file
    Reader(data_type : str, data_path: str, variables: Union[str, List])
    reader = Reader(...)
    reader.get_data(variables=Union[List, str])
    reader.get_user_data(user_id=..)
    '''
    def __init__(self, data_type: str, data_path: str, variables: Union[str, List]):
        '''
        Parameters : data_type : str, data_path: str, variables: Union[str, List]
        data_type : str
            Type of data to be read
            Only supports csv for now
        data_path : str
            Path to the data directory
        variables : Union[str, List]
            List of variables to be read
        '''
        self.data_type = data_type
        self.data_path = data_path
        # check if variables is a str
        # If so, convert it to a list
        if isinstance(variables, str):
            variables = [variables]
        self.variables = variables
        config_dict = {"local_directory": self.data_path}
        # check if data_type is csv
        if self.data_type == 'csv':
            self.reader_class = SparkCSVDataReader(config_dict, self.variables)
        else:
            raise NotImplementedError("Only csv data type is supported for now")

    def read_data(self):
        self.data = self.reader_class.read_data()

    def get_data(self, variables: Union[List, str]) -> RadarData:
        return self.data.get_combined_data_by_variable(variables)

    def get_user_data(self, user_id: str) -> RadarData:
        return self.data.get_data_by_user_id(user_id)



from abc import ABC, abstractmethod
from typing import List, Tuple

from radarpipeline.datalib import RadarData
from radarpipeline.datatypes import DataType
from radarpipeline.features.feature import Feature

import logging
logger = logging.getLogger(__name__)


class FeatureGroup(ABC):
    """
    A class to hold a group of features.
    """

    name: str
    description: str
    features: List["Feature"]
    required_input_data: List[str]

    def __init__(self, name: str, description: str, features: List["Feature"]):
        self.name = name
        self.description = description
        self.features = [f() for f in features]
        self.required_input_data = self._compute_required_data()

    def __str__(self):
        return self.name

    def get_required_data(self) -> List[str]:
        return self.required_input_data

    def get_listed_required_data(self, required_features: List[str]) -> List[str]:
        required_input_data = set()
        for feature in self.features:
            if feature.name in required_features:
                required_input_data.update(feature.get_required_data())
        return list(required_input_data)

    def _compute_required_data(self) -> List[str]:
        """
        Compute the required data for each feature in the group.
        """

        required_input_data = set()
        for feature in self.features:
            required_input_data.update(feature.get_required_data())
        return list(required_input_data)

    @abstractmethod
    def preprocess(self, data: RadarData) -> DataType:
        """
        Preprocess the data for each feature in the group.
        If there's nothing to process, please return the input
        """
        pass

    def get_all_features(self, data: RadarData) -> Tuple[List[str], List[DataType]]:
        """
        Compute the features for each feature in the group.
        """

        feature_names = []
        feature_values = []
        preprocessed_data = self.preprocess(data)
        for feature in self.features:
            print(feature.name)
            feature_names.append(feature.name)
            preprocessed_feature = feature.preprocess(preprocessed_data)
            feature_values.append(feature.calculate(preprocessed_feature))
        return feature_names, feature_values

    def get_listed_features(self, required_features: List[str],
                            data: RadarData) -> Tuple[List[str], List[DataType]]:
        """
        Compute the listed features for each feature in the group.
        """
        feature_names = []
        feature_values = []
        preprocessed_data = self.preprocess(data)
        for feature in self.features:
            if feature.name in required_features:
                feature_names.append(feature.name)
                preprocessed_feature = feature.preprocess(preprocessed_data)
                feature_values.append(feature.calculate(preprocessed_feature))
        # Check if all the required_features name are in feature_names.
        # if not, log warn the error and return the rest of the features
        if len(feature_names) != len(required_features):
            logger.warn("The required_features are not in the feature_names")
            # List the features that are not in the feature_names
            for feature in required_features:
                if feature not in feature_names:
                    logger.warn(f"{feature} is not in the part of the feature group")
        return feature_names, feature_values



from typing import List

import pandas as pd
import pyspark.sql.functions as f

from radarpipeline.datalib.abc import Data
from radarpipeline.datatypes import DataType


class RadarVariableData(Data):
    """
    Class for reading data for a single variable of a single user
    """

    _data: DataType

    def __init__(self, data: DataType, df_type: str = "pandas") -> None:
        self._data = data
        self.df_type = df_type
        self._preprocess_data()

    def get_data(self) -> DataType:
        return self._data

    def set_data(self, data: DataType) -> None:
        self._data = data

    def get_data_keys(self) -> List[str]:
        return list(self._data.columns)

    def get_data_size(self) -> int:
        if self.df_type == "pandas":
            return len(self._data.index)
        else:
            return int(self._data.count())

    def _preprocess_data(self) -> None:
        """
        Converts all time value columns to datetime format
        """

        if self.df_type == "spark":
            if "value.time" in self.get_data_keys():
                self._data = self._data.withColumn(
                    "value.time", f.to_date(self._data["`value.time`"])
                )
            if "value.timeReceived" in self.get_data_keys():
                self._data = self._data.withColumn(
                    "value.timeReceived", f.to_date(self._data["`value.timeReceived`"])
                )
            if "value.dateTime" in self.get_data_keys():
                self._data = self._data.withColumn(
                    "value.dateTime", f.to_date(self._data["`value.dateTime`"])
                )
        else:
            if "value.time" in self.get_data_keys():
                self._data["value.time"] = pd.to_datetime(
                    self._data["value.time"], unit="s"
                )
            if "value.timeReceived" in self.get_data_keys():
                self._data["value.timeReceived"] = pd.to_datetime(
                    self._data["value.timeReceived"], unit="s"
                )
            if "value.dateTime" in self.get_data_keys():
                self._data["value.dateTime"] = pd.to_datetime(
                    self._data["value.dateTime"], unit="s"
                )



from .mockfeatures.features import (
    MockFeatureGroup,
    PhoneBatteryChargingDuration,
    StepCountPerDay,
)

from .mockfeatures.features2 import (
    MockFeatureGroup2,
    StepCountPerHour
)


from typing import Dict, List, Optional, Union

import pandas as pd

from radarpipeline.common import utils
from radarpipeline.datalib.abc import Data
from radarpipeline.datalib.radar_user_data import RadarUserData
from radarpipeline.datatypes import DataType


class RadarData(Data):
    """
    Class for reading the RADAR data
    """

    _data: Dict[str, RadarUserData]

    def __init__(self, data: Dict[str, RadarUserData], df_type: str = "pandas") -> None:
        self._data = data
        self.df_type = df_type

    def get_data(self) -> Dict[str, RadarUserData]:
        return self._data

    def set_data(self, data: Dict[str, RadarUserData]) -> None:
        self._data = data

    def get_data_type(self) -> str:
        return self.df_type

    def get_data_keys(self) -> List[str]:
        return list(self._data.keys())

    def get_data_size(self) -> int:
        return len(self._data.items())

    def _get_data_by_key(self, key: str) -> Optional[RadarUserData]:
        return self._data.get(key, None)

    def _get_all_user_ids(self) -> List[str]:
        """
        Get all user ids in the data of the RADAR data

        Returns
        -------
        List[str]
            The list of all user ids in the data of the RADAR data
        """

        return self.get_data_keys()

    def get_combined_data_by_variable(
        self, variables: Union[str, List[str]]
    ) -> Union[DataType, List[DataType]]:
        """
        Returns the combined data of the RADAR data for the given variables

        Parameters
        ----------
        variables : Union[str, List[str]]
            The variable(s) to get the data for

        Returns
        -------
        Union[
            DataType,
            List[DataType]
        ]
            The combined data of the RADAR data for the given variables
        """

        is_only_one_var = False
        if isinstance(variables, str):
            is_only_one_var = True
            variables = [variables]

        all_user_ids = self._get_all_user_ids()
        variable_data_list = []
        variable_dict = {}

        # Store the data of the given variables of each user in a dictionary
        for user_id in all_user_ids:
            user_data = self._get_data_by_key(user_id)
            if user_data is not None:
                user_variables = user_data._get_all_variables()
                for var in variables:
                    if var in user_variables:
                        var_data = user_data.get_data_by_variable(var)
                        if var_data is not None:
                            if var not in variable_dict:
                                variable_dict[var] = []
                            variable_dict[var].append(var_data.get_data())

        # Combine the all data for each variable
        for var in variable_dict:
            if len(variable_dict[var]) > 0:
                if self.df_type == "spark":
                    combined_df = utils.combine_pyspark_dfs(variable_dict[var])
                else:
                    combined_df = pd.concat(variable_dict[var], ignore_index=True)
                variable_data_list.append(combined_df)

        if is_only_one_var:
            return variable_data_list[0]
        else:
            return variable_data_list

    def get_data_by_user_id(
        self, user_ids: Union[str, List[str]]
    ) -> Union[RadarUserData, List[RadarUserData]]:
        """
        Returns the data of the RADAR data for the given user ids

        Parameters
        ----------
        user_ids : Union[str, List[str]]
            The user id(s) to get the data for

        Returns
        -------
        Union[
            RadarUserData,
            List[RadarUserData]
        ]
            The data of the RADAR data for the given user ids
        """

        is_only_one_var = False
        if isinstance(user_ids, str):
            is_only_one_var = True
            user_ids = [user_ids]

        all_user_ids = self._get_all_user_ids()
        user_data_list = []

        for user_id in user_ids:
            if user_id in all_user_ids:
                user_data = self._get_data_by_key(user_id)
                if user_data is not None:
                    user_data_list.append(user_data)

        if is_only_one_var:
            return user_data_list[0]
        else:
            return user_data_list

    def get_variable_data(
            self, variables: Union[List, str]) -> Union[DataType, List[DataType]]:
        return self.get_combined_data_by_variable(variables)

    def get_user_data(self, user_id: str) -> Union[DataType, List[DataType]]:
        return self.get_data_by_user_id(user_id)



import json
import logging
import os
from typing import Any, Dict, List, Optional, Union
from multiprocessing import Pool
from functools import partial
from datetime import datetime
from radarpipeline.io.connection import SftpConnector

logger = logging.getLogger(__name__)


class SftpDataReader():
    """
    Class for reading data from SFTP
    """
    def __init__(self, config_dict: Dict, variables: List[str]):
        """
        Parameters
        ----------
        config_dict : Dict
            Dictionary containing configuration for SFTP
        variables : List
            List of variables to read
        root_dir : str, optional
        """

        self.config_dict = config_dict
        self.variables = variables
        if "sftp_target_path" not in config_dict:
            now = datetime.now()
            current_time = now.strftime("%d-%m-%y::%H:%M:%S")
            self.root_dir = os.path.join("tmp", current_time)
        else:
            self.root_dir = config_dict["sftp_target_path"]

    def read_sftp_data(self) -> None:
        sftp_connection_args = {}
        sftp_source_path = self.config_dict.get('sftp_source_path')
        sftp_connection_args["username"] = self.config_dict.get('sftp_username')
        sftp_connection_args["host"] = self.config_dict.get('sftp_host')
        sftp_connection_args["private_key"] = self.config_dict.get('sftp_private_key')
        all_participants_ids = self._get_all_id_sftp(sftp_source_path)
        func = partial(self._fetch_data, self.root_dir, sftp_source_path,
                       self.variables)
        try:
            func = partial(self._fetch_data, self.root_dir, sftp_source_path,
                           self.variables)
            with Pool(os.cpu_count()) as p:
                p.map(func, all_participants_ids)
        except Exception as e:
            logger.warn(f"Cannot use parallel processing to download the data from \
                sftp.Error: {e}")
            logger.warn("Downloading the data from sftp sequentially. \
                        This may take a while...")
            for uid in all_participants_ids:
                self._fetch_data(self.root_dir, sftp_source_path, self.variables, uid)
        logger.info(f"Data read from sftp and stored in {self.root_dir} folder")
        logger.info("To avoid redownloading, change config file to read from local")

    def get_root_dir(self) -> str:
        return self.root_dir

    def _fetch_data(self, root_path, sftp_source_path, included_var_cat, uid):
        sftp = SftpConnector(self.config_dict, included_var_cat)
        sftp.connect()
        try:
            with sftp.cd(os.path.join(sftp_source_path, uid)):
                source_folders = sftp.listdir(".")
                for src in source_folders:
                    if self._is_src_in_category(src, included_var_cat):
                        dir_path = os.path.join(uid, src)
                        os.makedirs(os.path.join(root_path, dir_path),
                                    exist_ok=True)
                        try:
                            with sftp.cd(src):
                                # To add new files only
                                src_files = sftp.listdir(".")
                                for src_file in src_files:
                                    # check if src file is a file or directory
                                    if sftp.isfile(src_file):
                                        sftp.get(src_file,
                                                 os.path.join(
                                                     root_path, dir_path, src_file),
                                                 preserve_mtime=True)
                                    else:
                                        if not os.path.exists(
                                            os.path.join(
                                                root_path, dir_path, src_file
                                            )
                                        ):
                                            os.makedirs(
                                                os.path.join(
                                                    root_path, dir_path,
                                                    src_file),
                                                exist_ok=True)
                                            sftp.get_d(src_file,
                                                       os.path.join(
                                                           root_path,
                                                           dir_path,
                                                           src_file),
                                                       preserve_mtime=True)
                        except FileNotFoundError:
                            print("Folder not found: " + dir_path + "/" + src_file)
                            continue
                        except EOFError:
                            print("EOFError: " + dir_path + "/" + src_file)
                            continue
        except FileNotFoundError:
            print("Folder not found: " + uid)
            return
        sftp.close()

    def _is_src_in_category(self, src, categories):
        if categories == "all":
            return True
        for category in categories:
            if src[:len(category)] == category:
                return True
        return False

    def _get_all_id_sftp(self, sftp_source_path):
        sftp = SftpConnector(self.config_dict, self.variables)
        sftp.connect()
        with sftp.cd(sftp_source_path + '/'):
            ids = [x for x in sftp.listdir() if x[0] != "."]
        sftp.close()
        return ids



import logging
import os
from typing import Dict, Optional

import pandas as pd
import pyspark.sql as ps

from radarpipeline.common import constants, utils
from radarpipeline.datatypes.data_types import DataType
from radarpipeline.io.abc import DataWriter

logger = logging.getLogger(__name__)


class SparkDataWriter(DataWriter):
    """
    Writes the data to a local directory using pySpark
    """

    features: Dict[str, ps.DataFrame]
    num_files: Optional[int] = None
    compression: str

    def __init__(
        self,
        features: Dict[str, DataType],
        output_dir: str,
        num_files: Optional[int] = 1,
        compress: bool = False,
    ) -> None:
        super().__init__(features, output_dir)
        self.num_files = num_files
        self.compression = "gzip" if compress is True else "none"

    def write_data(self) -> None:
        for feature_name, feature_df in self.features.items():
            folder_name = utils.pascal_to_snake_case(feature_name)
            folder_path = os.path.join(self.output_dir, folder_name)
            try:
                feature_df.write.csv(
                    path=folder_path,
                    header=True,
                    sep=constants.CSV_DELIMITER,
                    encoding=constants.ENCODING,
                    compression=self.compression,
                    lineSep=constants.LINESEP,
                )
            except Exception as e:
                logger.error(
                    f"Error writing data to file {folder_path}: {e}", exc_info=True
                )
            logger.info(f"Feature {feature_name} data exported to {folder_path}")


class PandasDataWriter(DataWriter):
    """
    Writes the data to a local directory using pandas
    """

    compression: str
    features: Dict[str, pd.DataFrame]

    def __init__(
        self,
        features: Dict[str, DataType],
        output_dir: str,
        compress: bool = False,
    ) -> None:
        super().__init__(features, output_dir)
        self.compression = "gzip" if compress is True else "infer"

    def write_data(self) -> None:
        for feature_name, feature_df in self.features.items():
            file_name = utils.pascal_to_snake_case(feature_name) + ".csv"
            file_name += ".gz" if self.compression == "gzip" else ""
            file_path = os.path.join(self.output_dir, file_name)
            try:
                feature_df.to_csv(
                    file_path,
                    index=False,
                    sep=constants.CSV_DELIMITER,
                    encoding=constants.ENCODING,
                    line_terminator=constants.LINESEP,
                    compression=self.compression,
                )
            except Exception as e:
                logger.error(
                    f"Error writing data to file {file_path}: {e}", exc_info=True
                )
            logger.info(f"Feature {feature_name} data exported to {file_path}")



from typing import Dict, List, Optional, Union

from radarpipeline.datalib.abc import Data
from radarpipeline.datalib.radar_variable_data import RadarVariableData


class RadarUserData(Data):
    """
    Class for reading data of a single user
    """

    _data: Dict[str, RadarVariableData]

    def __init__(
        self, data: Dict[str, RadarVariableData], df_type: str = "pandas"
    ) -> None:
        self._data = data
        self.df_type = df_type

    def get_data(self) -> Dict[str, RadarVariableData]:
        return self._data

    def set_data(self, data: Dict[str, RadarVariableData]) -> None:
        self._data = data

    def get_data_keys(self) -> List[str]:
        return list(self._data.keys())

    def get_data_size(self) -> int:
        return len(self._data.items())

    def _get_data_by_key(self, key: str) -> Optional[RadarVariableData]:
        return self._data.get(key, None)

    def _get_all_variables(self) -> List[str]:
        """
        Get all variables in the data of the user

        Returns
        -------
        List[str]
            The list of all variables in the data of the user
        """

        return self.get_data_keys()

    def get_data_by_variable(
        self, variables: Union[str, List[str]]
    ) -> Union[RadarVariableData, List[RadarVariableData]]:
        """
        Returns the data of the user for the given variables

        Parameters
        ----------
        variables : Union[str, List[str]]
            The variable(s) to get the data for

        Returns
        -------
        Union[
            RadarVariableData,
            List[RadarVariableData],
        ]
            The data of the user for the given variables
        """

        is_only_one_var = False
        if isinstance(variables, str):
            is_only_one_var = True
            variables = [variables]

        all_variables = self._get_all_variables()
        variable_data_list = []

        for var in variables:
            if var in all_variables:
                var_data = self._get_data_by_key(var)
                if var_data is not None:
                    variable_data_list.append(var_data)

        if is_only_one_var:
            return variable_data_list[0]
        else:
            return variable_data_list



from radarpipeline.io.abc import DataReader, SchemaReader
from radarpipeline.io.reader import AvroSchemaReader, SparkCSVDataReader, Reader
from radarpipeline.io.downloader import SftpDataReader
from radarpipeline.io.writer import *



from radarpipeline.datalib.abc import Data
from radarpipeline.datalib.radar_data import RadarData
from radarpipeline.datalib.radar_user_data import RadarUserData
from radarpipeline.datalib.radar_variable_data import RadarVariableData



import logging
import sys
import traceback

from radarpipeline import Project
from radarpipeline.common.logger import logger_init

logger_init()

logger = logging.getLogger(__name__)


def run(config_path: str = "config.yaml"):
    """
    Pipeline entry point.
    config_path could be a local path to a configuration file
    or a remote url to a configuration file.
    """
    try:
        logger.info("Starting the pipeline run...")
        logger.info("Reading and Validating the configuration file...")
        project = Project(input_data=config_path)
        logger.info("Fetching the data...")
        project.fetch_data()
        logger.info("Computing the features...")
        project.compute_features()
        logger.info("Exporting the features data...")
        project.export_data()
        logger.info("Pipeline run completed successfully")
    except KeyboardInterrupt:
        logger.info("Pipeline run interrupted by user")
        sys.exit(0)
    except Exception:
        logger.info(traceback.format_exc())
        sys.exit(1)


def _gather_clone_urls(organization, no_forks=True):
    """
    gh = pygithub3.Github()
    all_repos = gh.repos.list(user=organization).all()
    for repo in all_repos:

        # Don't print the urls for repos that are forks.
        if no_forks and repo.fork:
            continue

        yield repo.clone_url
    TODO: implement this function
    """


def show_available_pipelines():
    """
    Uses git to show the available pipelines from
    https://github.com/RADAR-base-Analytics/
    """
    """
    # Still in progress. not working yet
    clone_urls = _gather_clone_urls("RADAR-base-Analytics")
    print(clone_urls)
    return clone_urls
    # read all the repos from https://github.com/RADAR-base-Analytics/
    """


if __name__ == "__main__":
    run()



from abc import ABC, abstractmethod
from typing import List

from radarpipeline.datatypes.data_types import DataType


class Feature(ABC):
    name: str
    description: str
    required_input_data: List[str]

    def __init__(
        self, name: str, description: str, required_input_data: List[str]
    ) -> None:
        self.name = name
        self.description = description
        self.required_input_data = required_input_data

    def __str__(self) -> str:
        return self.name

    def get_required_data(self) -> List[str]:
        return self.required_input_data

    @abstractmethod
    def preprocess(self, data: DataType) -> DataType:
        """
        Preprocess the data for tje feature.
        If there's nothing to preprocess, please return the input
        """
        pass

    @abstractmethod
    def calculate(self, data: DataType) -> DataType:
        """
        Calculates the feature.
        """
        pass



from pyspark.sql.types import (
    BooleanType,
    ByteType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
)

ENCODING = "utf-8"
LINESEP = "\n"
CSV_DELIMITER = ","

STRING_TYPE = StringType()
BOOLEAN_TYPE = BooleanType()
BYTE_TYPE = ByteType()
SHORT_TYPE = ShortType()
INT_TYPE = IntegerType()
LONG_TYPE = LongType()
FLOAT_TYPE = FloatType()
DOUBLE_TYPE = DoubleType()

DATA_TYPE_MAPPING = {
    "byte": BYTE_TYPE,
    "short": SHORT_TYPE,
    "int": INT_TYPE,
    "long": LONG_TYPE,
    "float": FLOAT_TYPE,
    "double": DOUBLE_TYPE,
    "string": STRING_TYPE,
    "char": STRING_TYPE,
    "enum": STRING_TYPE,
    "array": STRING_TYPE,
    "boolean": BOOLEAN_TYPE,
}

INTEGER_TYPES = set([BYTE_TYPE, SHORT_TYPE, INT_TYPE, LONG_TYPE])
FLOATING_TYPES = set([FLOAT_TYPE, DOUBLE_TYPE])



#!/usr/bin/env python3
from setuptools import find_packages, setup
import os


def read_file(filename):
    with open(os.path.join(os.path.dirname(__file__), filename)) as file:
        return file.read()


setup(
    name="radarpipeline",
    version="2.0.1a",
    license='Apache',
    description="A python feature generation and visualization package use with RADAR project data.",
    url="https://github.com/RADAR-base/radarpipeline",
    download_url='https://github.com/RADAR-base/radarpipeline/archive/refs/tags/v2.0.0.tar.gz',
    readme="README.md",
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author="Heet Sankesara",
    author_email="heet.sankesara@kcl.ac.uk",
    keywords=['mhealth', 'pipeline', 'big-data'],
    packages=find_packages(),
    install_requires=[
        "pyYaml==6.0",
        "pandas==1.4.1",
        "numpy==1.22.3",
        "scipy==1.10.0",
        "pyspark[sql]==3.3.0",
        "GitPython==3.1.30"],
    test_suite="tests",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



import os
import logging
from git.exc import GitCommandError
from git.repo import Repo

from radarpipeline.common import utils

logger = logging.getLogger(__name__)


class ConfigValidator():
    def __init__(self, config, valid_input_formats, valid_output_formats) -> None:
        self.valid_input_formats = valid_input_formats
        self.valid_output_formats = valid_output_formats
        self.config = config

    def validate(self):
        """
        Validates the config to check if all the keys are present
        """

        required_keys = ["project", "input", "features", "output"]
        for key in required_keys:
            if key not in self.config:
                raise ValueError(f"Key not present in the config: {key}")

        self._validate_input()
        self._validate_configurations()
        self._validate_features()
        self._validate_output()

        logger.info("Config file validated successfully")

    def _update_mock_data(self) -> None:
        """
        Updates the mock data submodule of the åproject
        """

        try:
            MOCK_URL = "https://github.com/RADAR-base-Analytics/mockdata"
            cache_dir = os.path.join(
                os.path.expanduser("~"), ".cache", "radarpipeline", "mockdata")
            if not os.path.exists(cache_dir):
                logger.info("Mock data submodule not found. Cloning it...")
                Repo.clone_from(MOCK_URL, cache_dir)
                logger.info("Mock data input cloned")
            else:
                repo = Repo(
                    cache_dir,
                    search_parent_directories=True,
                )
                repo.remotes.origin.pull()

        except IndexError:
            raise ValueError("mockdata submodule not found")

    def _validate_input(self) -> None:
        """
        Validates the input data config
        """

        if self.config["input"]["data_type"] == "sftp":
            sftp_config_keys = [
                "sftp_host",
                "sftp_username",
                "sftp_source_path",
                "sftp_private_key",
            ]
            for key in sftp_config_keys:
                if key not in self.config["input"]["config"]:
                    raise ValueError(f"Key not present in the config: {key}")

        elif self.config["input"]["data_type"] == "local":
            if "source_path" not in self.config["input"]["config"]:
                raise ValueError("Key not present in the config: source_path")
            else:
                # Check if local_directory is absolute path. If not, then set it.
                local_directory = self.config["input"]["config"]["source_path"]
                if isinstance(local_directory, list):
                    local_directory = [
                        self._get_absolute_path(local_path)
                        for local_path in local_directory]
                    for local_path in local_directory:
                        if not os.path.exists(local_path):
                            raise ValueError(f"Path does not exist: {local_path}")
                else:
                    local_directory = utils.get_absolute_path(local_directory)
                    if not os.path.exists(local_directory):
                        raise ValueError(f"Path does not exist: {local_directory}")
                self.config["input"]["local_directory"] = local_directory

            # Raise error if data_format it not valid input formats
            if self.config["input"]["data_format"] not in self.valid_input_formats:
                raise ValueError("Invalid value for key in input: data_format")

        elif self.config["input"]["data_type"] == "mock":
            self._update_mock_data()

        else:
            raise ValueError("Invalid value for the key: data_location")

    def _validate_configurations(self) -> None:
        """
        Validates the configurations config
        """

        if "configurations" not in self.config:
            self.config["configurations"] = {}

        valid_df_types = ["pandas", "spark"]

        if "df_type" not in self.config["configurations"]:
            self.config["configurations"]["df_type"] = "pandas"
        elif self.config["configurations"]["df_type"] not in valid_df_types:
            raise ValueError("Invalid value for the key: df_type")

    def _validate_features(self) -> None:
        """
        Validates the features config
        """

        if len(self.config["features"]) == 0:
            raise ValueError("features array cannot be empty")

        for index, feature in enumerate(self.config["features"]):
            if "location" not in feature:
                raise ValueError(
                    f"Key not present in the config: location at index {index}"
                )

            if "feature_groups" not in feature:
                raise ValueError(
                    f"Key not present in the config: feature_groups at index {index}"
                )

            if len(feature["feature_groups"]) == 0:
                raise ValueError(
                    f"feature_groups array cannot be empty at index {index}"
                )
            if "feature_names" not in feature:
                raise ValueError(
                    f"Key not present in the config: feature_groups at index {index}"
                )

            if len(feature["feature_names"]) == 0:
                raise ValueError(
                    f"feature_groups array cannot be empty at index {index}"
                )
            feature_location = feature["location"]

            if utils.is_valid_github_path(feature_location):
                repo_name = utils.get_repo_name_from_url(feature_location)
                cache_dir = os.path.join(
                    os.path.expanduser("~"), ".cache", "radarpipeline", repo_name
                )

                if not os.path.exists(cache_dir):
                    Repo.clone_from(feature_location, cache_dir)

                repo = Repo(cache_dir)
                repo.git.reset("--hard")
                repo.git.clean("-xdf")

                active_branch_name = repo.active_branch.name
                feature_branch = feature.get("branch", active_branch_name)

                try:
                    repo.git.checkout(feature_branch)
                except GitCommandError:
                    logger.warning(
                        "Branch %s does not exist. Using the %s branch instead.",
                        feature_branch,
                        repo.active_branch.name,
                    )
                    feature_branch = repo.active_branch.name
                    repo.git.checkout(feature_branch)
                repo.remotes.origin.pull(feature_branch)

                feature_location = cache_dir
                logger.info(f"Using feature from cache location: {feature_location}")
            else:
                feature_location = os.path.expanduser(feature_location)
                if not os.path.isdir(feature_location):
                    raise ValueError(f"Invalid feature location: {feature_location}")
                logger.info(f"Using feature from local path: {feature_location}")

            self.config["features"][index]["location"] = feature_location

    def _validate_output(self) -> None:
        """
        Validates the output data config
        """

        if self.config["output"]["output_location"] == "postgres":
            postgres_config_keys = [
                "postgres_host",
                "postgres_username",
                "postgres_database",
                "postgres_password",
                "postgres_table",
            ]
            for key in postgres_config_keys:
                if key not in self.config["output"]['config']:
                    raise ValueError(f"Key not present in the config: {key}")

        elif self.config["output"]["output_location"] == "local":
            if "target_path" not in self.config["output"]['config']:
                raise ValueError("Key not present in the config: target_path")
            else:
                # Check if local_directory is absolute path. If not, then set it.
                local_directory = self.config["output"]['config']["target_path"]
                local_directory = utils.get_absolute_path(local_directory)
                if not os.path.exists(local_directory):
                    os.makedirs(local_directory, exist_ok=True)
                self.config["output"]['config']["target_path"] = local_directory

            # Raise error if data_format it not valid output formats
            if (
                self.config["output"]["data_format"]
                not in self.valid_output_formats
            ):
                raise ValueError("Invalid value for key in output: data_format")

            if "compress" not in self.config["output"]:
                self.config["output"]["compress"] = False
            if self.config["output"]["compress"] == "true":
                self.config["output"]["compress"] = True
        else:
            raise ValueError("Key not present in the config: output_location")
